#ifdef BEGIN_MAPPER
#include "mappers.hh" // SECOND INCLUSION

map<unsigned, mapper *>mappers;

BEGIN_MAPPER(0, )
	virtual void write(fWORD ofs, fBYTE value)
	{
        fprintf(stderr, "Unknown mapper wrote %02X to %04ZX\n", value, ofs);
    }
END_MAPPER(0)

/* MMC1 */
BEGIN_MAPPER(1, pos=reg=0;memset(regs,0,sizeof regs);regs[0]=0xC)
	BYTE reg;
	BYTE pos;
	BYTE regs[4];
	
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 1: [%04X] <- %02X\n", ofs, value);
		if(value&0x80)
			pos=reg=0;
		else
		{
			reg |= (value&1) << pos;
			if(++pos==5)
			{
				if(debug)fprintf(stderr, "Mapper 1: Reg value therefore %02X\n", reg);
				regs[ofs / 0x2000 - 4] = reg;
				rewrite();
				pos = reg = 0;
			}
		}
	}
	void rewrite()
	{
		reset();
		if(regs[0] & 8)
			maprom(2, (regs[0] & 4) ? 0 : 2, regs[3] & 15);
		else
			maprom(4, 0, regs[3] & 15);
		update();
	}
END_MAPPER(1)

/* UNROM */
BEGIN_MAPPER(2, )
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 2: [%04X] <- %02X\n", ofs, value);
		maprom(2, 0, value);
		update();
	}
END_MAPPER(2)

/* MMC4 */
BEGIN_MAPPER(4, )
	BYTE irqcounter, irqcountersub;
	BYTE cmd, prg0, prg1;
	bool prgswap, chrswap;
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 4: [%04X] <- %02X\n", ofs, value);
		
		switch(ofs & 0xE001)
		{
			case 0x8000:
				chrswap = value & 128;
				prgswap = value & 64;
				cmd = value & 7;
				exchange(0, 1, 0, 2, prgswap); // Swap $8000 and $C000 (8k)
				update();
				break;
			case 0x8001:
				switch(cmd)
				{
					case 6: maprom(1, 0, value); break; // Map 8k at $8000 (or $C000, if swapped)
					case 7: maprom(1, 1, value); break; // Map 8k at $A000
				}
				update();
				break;
			case 0xC000:
				// Countdown begins here
				irqcounter = value;
				//if(debug)
//				fprintf(stderr, "Mapper 4: IRQ counter %02X\n", value);
				break;
			case 0xC001:
				// Just load the counter
				irqcountersub = value;
				break;
			case 0xE000:
				// Reload the counter, don't activate
				irqcounter = irqcountersub;
				//if(debug)
//				fprintf(stderr, "Mapper 4: IRQ disabled\n");
				// disable irq
				break;
			case 0xE001:
				// enabled irq, countdown begins immediately
				//if(debug)
//				fprintf(stderr, "Mapper 4: IRQ enabled\n");
				break;
		}
	}
END_MAPPER(4)

BEGIN_MAPPER(6, )
	virtual void write(fWORD, fBYTE value)
	{
		maprom(2, 0, (value>>2) & 15);
		update();
	}
END_MAPPER(6)

BEGIN_MAPPER(7, )
	virtual void write(fWORD, fBYTE value)
	{
		maprom(4, 0, value & 15);
		update();
	}
END_MAPPER(7)

BEGIN_MAPPER(15, debug=true; maprom(4, 0, 0))
	virtual void write(fWORD ofs, fBYTE value)
	{
		if(debug)fprintf(stderr, "Mapper 15: [%04X] <- %02X\n", ofs, value);
		switch(ofs&3)
		{
			case 0:
			{
				unsigned i = value&63;
				maprom(2, 0, i);		// Select 16k at $8000
				maprom(2, 2, i+1 );		// Next 16k at $C000
				exchange(0, 1, 0, 1, !(value&128));
				exchange(1, 1, 2, 3, (value&128));
			//	exchange(0, 1, 0, 2, (value&128));
			//	exchange(1, 1, 1, 3, !(value&128));
				break;
			}
			case 1:
				maprom(2, 2, value&63);	// Select 16k at $C000
				exchange(1, 1, 2, 3, (value&128));
				break;
			case 2:
			{
				unsigned i = value;
				if(value >> 7)
				{
					++i;
					maprom(1, 1, i);
					maprom(1, 3, i);
				}
				else
				{
					maprom(1, 0, i);
					maprom(1, 2, i);
				}
				break;
			}
			case 3:
				/* &64 = mirroring control */
				maprom(2, 2, value&63);	// Select 16k at $C000
				exchange(1, 1, 2, 3, (value&128));
				break;
		}
		update();
	}
END_MAPPER(15)

#else
#error mappers.cc is not to be compiled individually!
#endif
