#include "cpu.hh"	/* DMC requires Op6502() */
#include "papu_internals.hh"

class dmcgen
{
	BYTE playbackmode;
	
	fWORD wavelen, wavepos;   // waveform position counter
    fWORD playaddr, loadaddr; // playing address
    fWORD dmalen, lengthleft; // playing length
public:
	bool irqflag;
    bool chanenabled;
    BYTE deltacounter;
private:
    inline void fetch()
    {
    	BYTE byt = Op6502(playaddr);
    	if(playaddr == 0xFFFF)playaddr = 0x8000;else ++playaddr;
    	
    	#if PAPU_DMC_DEBUG
    	fprintf(stderr, "%02X", byt);
    	#endif
    	
    	for(unsigned a=0; a<8; ++a)
    	{
    		if(byt & 1)
    		{
    			if(deltacounter < 126)deltacounter += 2;
    		}
    		else if(deltacounter > 1)deltacounter -= 2;
	    	byt >>= 1;
	    	//byt <<= 1;
	    }
	    #if PAPU_DMC_DEBUG
	    fprintf(stderr, "(%u)", deltacounter);
	    #endif
    }
public:
    dmcgen() : playbackmode(0), wavepos(0xFFF),
               playaddr(0xC000), loadaddr(0xC000), lengthleft(0),
               irqflag(false), chanenabled(false), deltacounter(0)
    {
    }
    void write(fBYTE ind, fBYTE value);

    inline void enable()
    {
        chanenabled = true;
        irqflag = false;
    }
    inline void disable()
    {
        chanenabled = false;
    }
    inline bool silent() const
    {
    	return !chanenabled || !lengthleft;
    }
    BEGIN_DMCWAVETICKER
    	fetch();    	
    	if(!--lengthleft)
    	{
    		if(playbackmode==2)
    		{
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC IRQ\n");
    			#endif
    			irqflag = true;
    			return true;
    		}
    		if(playbackmode & 1)
    		{
    			lengthleft = dmalen;
    			playaddr = loadaddr;
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC loop\n");
    			#endif
    		}
    		else
    		{
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC stop\n");
    			#endif
    		}
   		}
    END_DMCWAVETICKER
    
    void debug_silence()
    {
    	fprintf(stderr, "%sabled", chanenabled?"en":"dis");
    	fprintf(stderr, "; wlen=%u, mode=%d, counter=0x%02X, loading address=%X, length=%u, irqflag=%s",
    		wavelen, playbackmode, deltacounter, loadaddr, dmalen, irqflag?"set":"no");
    }
};
